@@ -17,6 +17,8 @@ import importlib
 import click
 import logging
 
+import odakb.config
+
 from os import getenv
 
 #import keyring
@@ -37,7 +39,7 @@ select = placeholder
 insert = placeholder
 
 
-logger = logging.getLogger("odakb.sparql")
+logger = logging.getLogger("oda.kb.sparql")
 
 def setup_logging(level=logging.INFO):
     handler = logging.StreamHandler(sys.stdout)
@@ -123,7 +125,7 @@ def load_defaults(default_prefixes, default_graphs):
                     logger.debug("appending new prefix: %s", p)
                     default_prefixes.append(p)
 
-            logger.info("\033[33mloaded prefixes from: %s\033[0m", odakb_defaults_http)
+            logger.debug("\033[33mloaded prefixes from: %s\033[0m", odakb_defaults_http)
 
             break
         except Exception as e:
@@ -141,7 +143,7 @@ def load_defaults(default_prefixes, default_graphs):
 
 def process_graph_loaders(G):
     for default_graph in LocalGraph.default_graphs:
-        logger.info("loading default graph %s", default_graph)
+        logger.debug("loading default graph %s", default_graph)
         load_graph(G, default_graph)
 
     q = """
@@ -156,7 +158,7 @@ def process_graph_loaders(G):
     for loader, rm, loc in G.query(q):
         logger.debug("loader: %s, rm: %s, loc: %s", loader, rm, loc)
         if str(rm) == "http://odahub.io/ontology#pythonModule":
-            logger.info("will load python module %s", loc)
+            logger.debug("will load python module %s", loc)
             mn, fn = loc.split(".")
             m = importlib.import_module(mn)
             getattr(m, fn)(G)
@@ -236,6 +238,8 @@ def get_jena_password():
 
     for n, m in {
                 'environ': lambda:os.environ['JENA_PASSWORD'].strip(),
+                'dynaconf': lambda:odakb.config.settings.jena_password,
+                'homefile': lambda:open(os.path.join(os.environ.get('HOME'), '.jena-password')).read().strip(),
                 #'keyring': lambda:keyring.get_password("jena", "admin"),
             }.items():
         try:
@@ -243,7 +247,7 @@ def get_jena_password():
             logger.info("good JENA password from %s", n)
             return r
         except Exception as e:
-            logging.warning("failed %s %s %s", n, m, e)
+            logger.warning("failed %s %s %s", n, m, e)
             tried.append([n, m, e])
 
     raise RuntimeError("no good jena password, tried: "+repr(tried))
@@ -259,11 +263,14 @@ def compose_sparql(body, prefixes=None):
 @cli.command("construct")
 @click.argument("data")
 @click.option("-j", "--jsonld", is_flag=True, default=False)
-def _construct(data, jsonld):
+@click.option("-o", "--output", default=None)
+def _construct(data, jsonld, output):
     r = construct(data, jsonld)
-
+            
     if jsonld:
         logger.debug(json.dumps(r, indent=2, sort_keys=True))
+        if output is not None and output.strip() == "-":
+            json.dump(r, sys.stdout, indent=2, sort_keys=True)
     else:
         logger.debug(r)
 
@@ -290,47 +297,62 @@ def _execute_sparql(data, endpoint="query", service=None):
     r = execute_sparql(compose_sparql(data), endpoint, invalid_raise=True, raw=False, service=service)
 
     for e in r['results']['bindings']:
-        logger.info("entry: %s", e)
+        logger.debug("entry: %s", e)
+
+
+def discover_oda_sparql_root(service):
+    oda_sparql_root = None
+    
+    for n, m, obsolete in [
+                ("service argument", lambda: service, False),
+                ("default_oda_sparql_root", lambda: default_oda_sparql_root, False),
+                ("HOME/.oda-sparql-root", lambda: open(os.path.join(getenv("HOME"), ".oda-sparql-root")).read().strip(), False),
+                ("dynaconf", lambda: odakb.config.settings.sparql_root, False),
+                ("default public endpoint", lambda: "https://www.astro.unige.ch/cdci/astrooda/dispatch-data/gw/odakb", False)
+            ]:
+        try:
+            oda_sparql_root = m()
+            if oda_sparql_root is None:
+                logger.debug("\033[33mfailed to discover oda_sparql_root with %s as it returns None\033[0m", n)
+            else:
+                logger.info("\033[32mdiscovered oda_sparql_root with %s: %s\033[0m", n, oda_sparql_root)
+                if obsolete:
+                    logger.warning("using %s, this will be discarded soon", n)
+                break
+        except Exception as e:
+            logger.debug("\033[33mfailed to discover oda_sparql_root with %s due to %s\033[0m", n, e)                                
+
+    return oda_sparql_root
+
 
 def execute_sparql(data, endpoint, invalid_raise, raw=False, service=None):
     logger.debug("data: %s", repr(data))
         
+    oda_sparql_root = discover_oda_sparql_root(service)
 
     t0=time.time()
 
-    if service is None:
-        if default_oda_sparql_root is not None:
-            oda_sparql_root = default_oda_sparql_root
-            logger.info("using default sparql endpoint from \033[32mdefault_oda_sparql_root\033[0m module variable")
-        else:
-            oda_sparql_root = os.environ.get("ODA_SPARQL_ROOT", )
-            if oda_sparql_root:
-                logger.info("using sparql endpoing from \033[32mODA_SPARQL_ROOT\033[0m environment variable")
-            else:
-                oda_sparql_root = "https://www.astro.unige.ch/cdci/astrooda/dispatch-data/gw/odakb"
-                logger.info("using default sparql endpoint")
-    else:
-        oda_sparql_root = service
-
     logger.info("ODA Knowledge Base (SPARQL) root is \033[32m%s\033[0m", oda_sparql_root)
 
+    url = oda_sparql_root + "/" + endpoint
+
     if endpoint == "update":    
         auth=requests.auth.HTTPBasicAuth("admin", 
                                          get_jena_password())
-        r=requests.post(oda_sparql_root+"/"+endpoint,
+        r=requests.post(url,
                         data=data,
                         auth=auth
                         )
     elif endpoint == "query":
         auth=None
-        r=requests.post(oda_sparql_root+"/"+endpoint,
-                       params=dict(query=data)
-                    )
+        r=requests.post(url,
+                        params=dict(query=data)
+                       )
     else:
         auth=None
         r=requests.post(oda_sparql_root,
-                       params=dict(query=data)
-                    )
+                        params=dict(query=data)
+                       )
 
     note_stats(spent_seconds=time.time()-t0, query_size=len(data))
     
@@ -348,13 +370,13 @@ def execute_sparql(data, endpoint, invalid_raise, raw=False, service=None):
             logger.debug("serice returns %s", r.text)
         
             if r.status_code == 403:
-                m = f"Acceess To ODA KB denied at {oda_sparql_root}\n"
+                m = f"Access To ODA KB denied at {oda_sparql_root}\n"
                 m += f"please refer to https://github.com/volodymyrss/oda-kb.git for more info "
                 logger.error(m)
                 raise SPARQLException(m,
-                                      r.status_code, r.text)
+                                      r.status_code, r.text, url)
             else:
-                raise SPARQLException(r.status_code, r.text)
+                raise SPARQLException(r.status_code, r.text, url)
 
     if raw:
         return r.text
@@ -447,7 +469,7 @@ def _select(query=None, form=None, todict=True, tojson=False, tordf=False, tojdi
         jsonld = g.serialize(format='json-ld', indent=4, sort_keys=True).decode()
 
     if tojson:
-        logger.debug(jsonld)
+        print(jsonld) # TODO: where?
         return jsonld
     
     if tojdict:
@@ -477,12 +499,12 @@ def _select(query=None, form=None, todict=True, tojson=False, tordf=False, tojdi
 
         jdict = jsonld2dict(json.loads(jsonld))
         
-        logger.debug(json.dumps(jdict, sort_keys=True, indent=4))
+        print(json.dumps(jdict, sort_keys=True, indent=4)) # TODO: where?
 
         return jdict
 
     for e in entries:
-        logger.info("found fact: %s", e)
+        logger.debug("found fact: %s", e)
 
     if todict:
         return entries 
